from housekeeper.store.models import File
from housekeeper.store.store import Store


class FileService:
    def __init__(self, store: Store):
        self.store = store

    def get_local_files(self, bundle: str, tags: list[str], version_id: int) -> list[File]:
        return self.store.get_files(
            bundle_name=bundle, tag_names=tags, version_id=version_id, local_only=True
        ).all()

    def get_remote_files(self, bundle: str, tags: list[str], version_id: int) -> list[File]:
        return self.store.get_files(
            bundle_name=bundle, tag_names=tags, version_id=version_id, remote_only=True
        ).all()



"""Tests for cli get file functionality"""

from pathlib import Path

from housekeeper.cli.get import files_cmd
from housekeeper.services.file_report_service.utils import _get_suffix, squash_names
from housekeeper.store.models import File
from housekeeper.store.store import Store


def test_get_files_no_files(base_context, cli_runner):
    """Test to get all files when there are no files"""
    # GIVEN a context with a populated store and a cli runner
    store = base_context["store"]

    # GIVEN a store without files
    assert not store.get_files().all()

    # WHEN fetching all files by not specifying any file
    result = cli_runner.invoke(files_cmd, ["--json"], obj=base_context)

    # THEN assert that output was produced
    assert result.exit_code == 0


def test_get_files_json(populated_context, cli_runner):
    """Test to get all files from a populated store"""
    # GIVEN a context
    store: Store = populated_context["store"]

    # GIVEN a store with files
    files: list[File] = store.get_files()

    # WHEN fetching all files
    result = cli_runner.invoke(files_cmd, ["--json"], obj=populated_context)

    # THEN assert that all files where printed
    for file in files:
        assert file.path in result.output


def test_get_files(populated_context, cli_runner):
    """Test to get all files from a populated store in human friendly format"""
    # GIVEN a context and a store
    store: Store = populated_context["store"]

    # GIVEN a store with some files
    assert store.get_files().all()

    # GIVEN a file name
    file: File = store.get_files().first()
    file_name: str = Path(file.path).name

    # WHEN fetching all files by not specifying any file
    result = cli_runner.invoke(files_cmd, [], obj=populated_context)

    # THEN assert that the file name is displayed
    assert file_name in result.output

    # THEN assert that the full file path is not shown
    assert file.path not in result.output


def test_get_files_tag(populated_context, cli_runner, vcf_tag_name):
    """Test to get files with a specific tag from a populated store"""
    # GIVEN a context with a populated store
    store: Store = populated_context["store"]

    # GIVEN a store with some files that are tagged
    tagged_files: list[File] = store.get_files(tag_names=[vcf_tag_name]).all()
    assert tagged_files

    # WHEN fetching all files by not specifying any file
    result = cli_runner.invoke(files_cmd, ["--tag", vcf_tag_name, "--json"], obj=populated_context)

    # THEN assert that all files where fetched
    for file in tagged_files:
        assert file.path in result.output


def test_get_files_multiple_tags(populated_context, cli_runner, vcf_tag_name, family_tag_name):
    """Test to get files with multiple tags tag from a populated store"""
    # GIVEN a context with a populated store
    store: Store = populated_context["store"]

    # GIVEN a store with some files that are tagged
    files: list[File] = store.get_files(tag_names=[vcf_tag_name, family_tag_name]).all()
    assert files

    # WHEN fetching all files by not specifying any file
    result = cli_runner.invoke(
        files_cmd,
        ["--tag", vcf_tag_name, "--tag", family_tag_name, "--json"],
        obj=populated_context,
    )

    # THEN assert that all files where fetched
    for file in files:
        assert file.path in result.output


def test_get_files_rare_tag(populated_context, cli_runner, family_tag_name):
    """Test to get files with a tag that is not on all files from a populated store"""
    # GIVEN a context with a populated store and a cli runner
    store: Store = populated_context["store"]

    # GIVEN a store with some tagged files
    tagged_files: list[File] = store.get_files(tag_names=[family_tag_name]).all()

    # WHEN fetching all tagged files
    result = cli_runner.invoke(
        files_cmd, ["--tag", family_tag_name, "--json"], obj=populated_context
    )

    # THEN all tagged files where fetched
    for tagged_file in tagged_files:
        assert tagged_file.path in result.output


def test_get_files_compact():
    """Test to get all files from a populated store in human friendly format, subsequent names concatenated"""
    # GIVEN an example result file list
    file_list = [
        {"path": "family.vcf", "full_path": "tests/family.vcf", "tags": [], "id": 7},
        {
            "path": "family.2.vcf",
            "full_path": "/tests/family.2.vcf",
            "tags": [],
            "id": 8,
        },
        {
            "path": "family.3.vcf",
            "full_path": "/tests/family.3.vcf",
            "tags": [],
            "id": 9,
        },
    ]

    # WHEN calling `squash_names` on list of files
    squashed = squash_names(file_list)

    # THEN assert that the file names displayed are squashed
    assert len(squashed) < len(file_list)


def test_get_suffix():
    # GIVEN
    dont_split_name1 = "asdf2.asdf.vcf"
    dont_split_name2 = "1123"
    dont_split_name3 = "asdf.vcf"
    dont_split_name4 = "asdf_2.asdf.vcf"

    split_name1 = "asdf1.png"
    split_name2 = "asdf8A8_7777_asdf_8.png"

    # THEN assert filename parsing *does not* split name into (prefix, integer, suffix)
    assert (dont_split_name1, "", "") == _get_suffix(dont_split_name1)
    assert (dont_split_name2, "", "") == _get_suffix(dont_split_name2)
    assert (dont_split_name3, "", "") == _get_suffix(dont_split_name3)
    assert (dont_split_name4, "", "") == _get_suffix(dont_split_name4)

    # THEN assert filename parsing *does* split name into (prefix, integer, suffix)
    assert (split_name1, "", "") != _get_suffix(split_name1)
    assert (split_name2, "", "") != _get_suffix(split_name2)



# -*- coding: utf-8 -*-
"""This is a core module for the Store API"""

import logging
from pathlib import Path

from housekeeper.store.crud.create import CreateHandler
from housekeeper.store.crud.read import ReadHandler
from housekeeper.store.crud.update import UpdateHandler
from housekeeper.store.database import get_session
from housekeeper.store.models import File, Version

LOG = logging.getLogger(__name__)


class Store(CreateHandler, ReadHandler, UpdateHandler):
    """
    Handles interactions with the database in the context when a temporary
    database connection is needed, e.g., a command line interface.
    """

    def __init__(self, root: str):
        self.session = get_session()
        ReadHandler(self.session)
        CreateHandler(self.session)
        UpdateHandler(self.session)

        LOG.debug("Initializing Store")
        File.app_root = Path(root)
        Version.app_root = Path(root)

        super().__init__(self.session)



"""Code for CLI get"""

import json as jsonlib
import logging

import click
from rich.console import Console

from housekeeper.services.file_service.file_service import FileService
from housekeeper.services.file_report_service.file_report_service import FileReportService
from housekeeper.store.api import schema
from housekeeper.store.models import Version
from housekeeper.store.store import Store

from .tables import (
    get_bundles_table,
    get_tags_table,
    get_versions_table,
)

LOG = logging.getLogger(__name__)


@click.group()
def get():
    """Get info from database"""


@get.command("bundle")
@click.argument("bundle-name", required=False)
@click.option("-i", "--bundle-id", type=int, help="Search for a bundle with bundle id")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option(
    "-c",
    "--compact",
    is_flag=True,
    help="print compact filenames IFF verobe flag present",
)
@click.pass_context
def bundle_cmd(context, bundle_name, bundle_id, json, compact):
    """Get bundle information from database"""
    store: Store = context.obj["store"]
    bundles = store.bundles()

    if bundle_name:
        bundle = store.get_bundle_by_name(bundle_name=bundle_name)
        bundles = [bundle] if bundle else []

    if bundle_id:
        bundle = store.get_bundle_by_id(bundle_id=bundle_id)
        bundles = [bundle] if bundle else []

    if not bundles:
        LOG.info("Could not find any bundles")
        return
    template = schema.BundleSchema()
    result = []
    for bundle in bundles:
        result.append(template.dump(bundle))

    if json:
        click.echo(jsonlib.dumps(result, indent=4, sort_keys=True))
        return
    console = Console()
    console.print(get_bundles_table(result))
    for bundle in bundles:
        if len(bundle.versions) == 0:
            LOG.info("No versions found for bundle %s", bundle.name)
            return
        version_obj = bundle.versions[0]
        context.invoke(version_cmd, version_id=version_obj.id, verbose=True, compact=compact)


@get.command("version")
@click.option("-b", "--bundle-name", help="Fetch all versions from a bundle")
@click.option("-i", "--version-id", type=int, help="Fetch a specific version")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-v", "--verbose", is_flag=True, help="print additional information")
@click.option(
    "-c",
    "--compact",
    is_flag=True,
    help="print compact filenames IFF verobe flag present",
)
@click.pass_context
def version_cmd(context, bundle_name, json, version_id, verbose, compact):
    """Get versions from database"""
    store: Store = context.obj["store"]
    if not (bundle_name or version_id):
        LOG.info("Please select a bundle or a version")
        return
    if bundle_name:
        bundle = store.get_bundle_by_name(bundle_name=bundle_name)
        if not bundle:
            LOG.info("Could not find bundle %s", bundle_name)
            return
        version_objs = bundle.versions

    if version_id:
        version: Version = store.get_version_by_id(version_id=version_id)
        if not version:
            LOG.warning("Could not find version %s", version_id)
            raise click.Abort
        version_objs = [version]

    version_template = schema.VersionSchema()
    result = []
    for version_obj in version_objs:
        bundle = store.get_bundle_by_id(bundle_id=version_obj.bundle_id)
        res = version_template.dump(version_obj)
        res["bundle_name"] = bundle.name
        result.append(res)

    if json:
        click.echo(jsonlib.dumps(result))
        return

    console = Console()
    console.print(get_versions_table(result))
    if not verbose:
        return

    for version_obj in version_objs:
        context.invoke(files_cmd, version_id=version_obj.id, verbose=True, compact=compact)


@get.command("file")
@click.option("-t", "--tag", "tag_names", multiple=True, help="filter by file tag")
@click.option("-v", "--version", "version_id", type=int, help="filter by version of the bundle")
@click.option("-V", "--verbose", is_flag=True, help="print additional information")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-c", "--compact", is_flag=True, help="print compact filenames")
@click.argument("bundle", required=False)
@click.pass_context
def files_cmd(
    context,
    tag_names: list[str],
    version_id: int,
    verbose: bool,
    bundle: str,
    json: bool,
    compact: bool,
):
    """Get files from database"""
    file_service: FileService = context.obj["file_service"]
    output_service: FileReportService = context.obj["file_report_service"]

    output_service.compact = compact
    output_service.json = json

    local = file_service.get_local_files(bundle=bundle, tags=tag_names, version_id=version_id)
    remote = file_service.get_remote_files(bundle=bundle, tags=tag_names, version_id=version_id)

    output_service.log_file_table(files=local, header="Local files", verbose=verbose)
    output_service.log_file_table(files=remote, header="Remote files", verbose=False)


@get.command("tag")
@click.option("-j", "--json", is_flag=True, help="Output to json format")
@click.option("-n", "--name", multiple=True, help="Specify a tag name")
@click.pass_context
def tag_cmd(context, json, name):
    """Get the tags from database"""
    store: Store = context.obj["store"]
    LOG.info("Fetch tags")
    tag_objs = store.get_tags()
    template = schema.TagSchema()
    result = []
    for tag_obj in tag_objs:
        if name and (tag_obj.name not in name):
            continue
        LOG.debug("Use tag %s", tag_obj.name)
        result.append(template.dump(tag_obj))
    if not result:
        LOG.info("Could not find any of the specified tags [%s]", ", ".join(name))
        return
    if json:
        click.echo(jsonlib.dumps(result))
        return
    console = Console()
    console.print(get_tags_table(result))



"""Code for building tables that are displayed in the terminal"""

import re
from pathlib import Path

from rich.table import Table


def get_tags_table(rows: list[dict]) -> Table:
    """Return a tag table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:bookmark:[/] Tags table [not italic]:bookmark:[/]"
    table.add_column("ID")
    table.add_column("Name")
    table.add_column("Category")
    table.add_column("Created")
    for tag_obj in rows:
        table.add_row(
            str(tag_obj["id"]),
            tag_obj["name"],
            tag_obj["category"],
            tag_obj["created_at"].split("T")[0],
        )
    return table


def get_bundles_table(rows: list[dict]) -> Table:
    """Return a bundles table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:package:[/] Bundle table [not italic]:package:[/]"
    table.add_column("ID")
    table.add_column("Bundle name")
    table.add_column("Version IDs")
    table.add_column("Created")
    for bundle_obj in rows:
        bundle_versions = ", ".join(str(version["id"]) for version in bundle_obj["versions"])
        table.add_row(
            str(bundle_obj["id"]),
            bundle_obj["name"],
            bundle_versions,
            bundle_obj["created_at"].split("T")[0],
        )
    return table


def get_versions_table(rows: list[dict]) -> Table:
    """Return a versions table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = "[not italic]:closed_book:[/] Version table [not italic]:closed_book:[/]"
    table.add_column("ID")
    table.add_column("Bundle name")
    table.add_column("Nr files")
    table.add_column("Included")
    table.add_column("Archived")
    table.add_column("Created")
    table.add_column("Expires")
    for version_obj in rows:
        table.add_row(
            str(version_obj["id"]),
            version_obj["bundle_name"],
            str(len(version_obj["files"])),
            version_obj["included_at"].split("T")[0] if version_obj["included_at"] else "",
            version_obj["archived_at"].split("T")[0] if version_obj["archived_at"] else "",
            version_obj["created_at"].split("T")[0] if version_obj["created_at"] else "",
            version_obj["expires_at"].split("T")[0] if version_obj["expires_at"] else "",
        )
    return table



"""Fixture for the Store tests"""

import datetime
import datetime as dt
from copy import deepcopy
from pathlib import Path

import pytest

from housekeeper.store import models
from housekeeper.store.models import Bundle, File, Version
from housekeeper.store.store import Store


@pytest.fixture(scope="function")
def minimal_bundle_obj(case_id, timestamp) -> models.Bundle:
    """Return a bundle object"""
    return models.Bundle(name=case_id, created_at=timestamp)


@pytest.fixture(scope="function")
def bundle_data_json(bundle_data):
    """Fixture for bundle data in json format"""
    json_data = deepcopy(bundle_data)
    json_data["created_at"] = json_data["created_at"].isoformat()
    return json_data


@pytest.fixture(scope="function")
def second_timestamp(timestamp) -> datetime.datetime:
    """Return a time stamp in date time format"""
    return timestamp + datetime.timedelta(days=10)


@pytest.fixture(scope="function")
def old_timestamp() -> datetime.datetime:
    """Return a time stamp that is older than a year"""
    return datetime.datetime(2018, 1, 1)


@pytest.fixture(scope="function")
def second_bundle_data(bundle_data, second_sample_vcf, second_family_vcf, second_timestamp) -> dict:
    """Return a bundle similar to bundle_data with updated file paths"""
    second_bundle = deepcopy(bundle_data)
    second_bundle["created_at"] = second_timestamp
    second_bundle["files"][0]["path"] = str(second_sample_vcf)
    second_bundle["files"][1]["path"] = str(second_family_vcf)
    return second_bundle


@pytest.fixture(scope="function")
def other_case() -> str:
    """Return the case id that differs from base fixture"""
    return "angrybird"


@pytest.fixture(scope="function")
def bundle_data_old(
    bundle_data, second_sample_vcf, second_family_vcf, old_timestamp, other_case
) -> dict:
    """Return info for a older bundle with different files and case id as bundle data"""
    _bundle = deepcopy(bundle_data)
    _bundle["name"] = other_case
    _bundle["created_at"] = old_timestamp
    _bundle["files"][0]["path"] = str(second_sample_vcf)
    _bundle["files"][1]["path"] = str(second_family_vcf)
    return _bundle


@pytest.fixture(scope="function")
def time_stamp_now() -> dt.datetime:
    """Time stamp now"""
    return dt.datetime.now()


@pytest.fixture
def store_for_testing_getting_archived_files(store: Store) -> Store:
    """Returns a store containing a bundle with four spring files, one not archived,
    one archived, one being retrieved and one which has been retrieved."""
    tag = store.new_tag(name="spring")
    store.session.add(tag)
    store.session.commit()

    bundle: Bundle = store.new_bundle(name="sample_id", created_at=datetime.datetime.now())
    version: Version = store.new_version(created_at=datetime.datetime.now())
    bundle.versions.append(version)
    not_archived_file: File = store.add_file(
        bundle=bundle, file_path=Path("not", "archived", "file.txt"), tags=["spring"]
    )
    archived_file: File = store.add_file(
        bundle=bundle, file_path=Path("archived", "file.txt"), tags=["spring"]
    )
    archived_file_ongoing_retrieval = store.add_file(
        bundle=bundle,
        file_path=Path("retrieval", "ongoing", "archived", "file.txt"),
        tags=["spring"],
    )
    retrieved_file = store.add_file(
        bundle=bundle, file_path=Path("retrieved", "file.txt"), tags=["spring"]
    )
    store.session.add(bundle)
    store.session.add(not_archived_file)
    store.session.add(archived_file)
    store.session.add(archived_file_ongoing_retrieval)
    store.session.add(retrieved_file)
    store.session.add(version)
    store.session.commit()
    archive_retrieval_not_ongoing = store.create_archive(
        file_id=archived_file.id, archiving_task_id=1
    )
    archive_retrieval_ongoing = store.create_archive(
        file_id=archived_file_ongoing_retrieval.id, archiving_task_id=2
    )
    archive_retrieved = store.create_archive(file_id=retrieved_file.id, archiving_task_id=3)
    archive_retrieval_ongoing.archived_at = datetime.datetime.now()
    archive_retrieval_ongoing.retrieval_task_id = 1
    archive_retrieved.archived_at = datetime.datetime.now()
    archive_retrieved.retrieval_task_id = 2
    archive_retrieved.retrieved_at = datetime.datetime.now()
    store.session.add(archive_retrieval_ongoing)
    store.session.add(archive_retrieval_not_ongoing)
    store.session.add(archive_retrieved)
    store.session.commit()
    return store



import re
from pathlib import Path

from rich.table import Table

from housekeeper.store.api import schema
from housekeeper.store.models import File


def format_files(files: list[File]):
    template = schema.FileSchema()
    formatted_files = []
    for file in files:
        formatted_file = template.dump(file)
        formatted_files.append(formatted_file)
    return formatted_files


def get_files_table(rows: list[dict], header: str, verbose=False, compact=False) -> Table:
    """Return a tag table"""
    table = Table(show_header=True, header_style="bold magenta")
    table.title = f"[not italic]:scroll:[/] {header} [not italic]:scroll:[/]"
    table.add_column("ID")
    table.add_column("File name")
    table.add_column("Tags")
    if compact:
        rows = squash_names(rows)
    for i, file_obj in enumerate(rows, 1):
        file_tags = ", ".join(tag["name"] for tag in file_obj["tags"])
        file_path = Path(file_obj["path"])
        file_name = file_path.name
        if verbose:
            file_name = file_obj["full_path"]

        if i % 2 == 0:
            table.add_row(str(file_obj["id"]), f"[yellow]{file_name}[/yellow]", file_tags)
        else:
            table.add_row(str(file_obj["id"]), f"[cyan]{file_name}[/cyan]", file_tags)
    return table


def squash_names(list_of_files: list[dict]) -> list[dict]:
    """If subsequent elements (filenames) in 'list_of_files' end in an integer - And that integer is
    following the previous - those are squashed when displayed.
    Example:

        ["asdf1.txt", "asdf2.txt"] becomes asdf[1-2].txt'

    Algorithm summary

        Traverse a list of dictionaries, where each dictionary represents
        a file with 'name', 'path' and 'tag', etc. For each name in the list ending
        with integer n, for every subsequent name ending in n+1, n+2, ... n+i displayed
        name will be: name[n-i].

        The input list will not be sorted.

        For each element e in the list, e will be searched by regular expression for integers
         at a specific location; this is cached for the next iteration where a comparison is done. If
        subsequent cache is updated, if not, the result is printed. Print will look in the cache
        to check if the current iteration has a subsequent integer ending. Tags associated with squashed
        filenames will be sorted and duplicates removed.
    """
    list_of_squashed = []
    tag_list = []
    if not list_of_files:
        return list_of_squashed
    head = list_of_files[0]
    tail = list_of_files[1:]
    (previous_file, previous_counter, previous_suffix) = _get_suffix(head["path"])
    squash = [previous_counter]
    previous_hkjson = head
    for hk_json in tail + [{"path": ""}]:  # Add padding for final iteration
        (filename, counter, suffix) = _get_suffix(hk_json["path"])
        if counter == str(_to_int(previous_counter) + 1) and (previous_file == filename):
            squash = squash + [counter]
            tag_list = tag_list + (hk_json["tags"])
        elif len(squash) == 1:  # only previous element in the list
            list_of_squashed.append(previous_hkjson)
            squash = [counter]
        else:
            squashed_path = f"{previous_file}[{squash[0]}-{squash[-1]}]{previous_suffix}"
            previous_hkjson["path"] = squashed_path
            previous_hkjson["full_path"] = squashed_path
            previous_hkjson["tags"] = remove_duplicates(sorted(tag_list, key=lambda i: i["id"]))
            previous_hkjson["id"] = "-"
            list_of_squashed.append(previous_hkjson)
            squash = [counter]
            tag_list = []
        previous_counter = counter
        previous_suffix = suffix
        previous_file = filename
        previous_hkjson = hk_json
    return list_of_squashed


def remove_duplicates(tag_list: list[dict]) -> list[dict]:
    """Remove duoplicate elements from `tag_list`"""
    no_duplicates = []
    for i in tag_list:
        if i not in no_duplicates:
            no_duplicates.append(i)
    return no_duplicates


def _get_suffix(filename: str) -> tuple[str, str, str]:
    """Split a filename if ending with an integer before suffix."""
    parsed: list[str] = re.split(pattern=r"(\d+)\.(\w{2,3}$)", string=filename)
    if len(parsed) == 4:
        return parsed[0], parsed[1], f".{parsed[2]}"
    return filename, "", ""


def _to_int(string):
    """Cast to int, empty string becomes 0"""
    return 0 if string == "" else int(string)
